/*-----------------------------------------------------------------*-C-*---
 * File:    handc/runtime/profile.c
 *
 *          Copyright (C)1997 Donovan Kolbly <d.kolbly@rosette.com>
 *          as part of the RScheme project, licensed for free use
 *
 * Version: 1.4
 * Date:    1997.01.23 09:08:07
 * Build:   v0.7-0.9, 97.01.26
 *
 * Purpose: User-level profiling support
 *------------------------------------------------------------------------*/

#include <rscheme/runtime.h>
#include <rscheme/vinsns.h>
#include <rscheme/hashmain.h>
#include <rscheme/osglue.h>

obj rsprof_object_tag( obj tmpl )
{
  obj str = bvec_alloc( sizeof(obj)+1, string_class );
  memcpy( PTR_TO_DATAPTR(str), &tmpl, sizeof(obj) );
  return str;
}

#if ACCUM_GF_CACHE_HISTOGRAM
int gf_cache_hit_depth[100];

obj collect_gf_cache_histogram( rs_bool reset_q )
{
  obj v = alloc( SLOT(100), vector_class );
  unsigned i;

  for (i=0; i<100; i++)
    gvec_write_init_non_ptr( v, SLOT(i), int2fx(gf_cache_hit_depth[i]) );

  if (reset_q)
    for (i=0; i<100; i++)
      gf_cache_hit_depth[i] = 0;

  return v;
}
#else
obj collect_gf_cache_histogram( rs_bool reset_q )
{
  return make0(vector_class);
}
#endif

#if !PROFILING_HOOKS

void rsprof_start( obj setup )
{
  scheme_error( "profiling not enabled", 0 );
}

obj rsprof_stop( void )
{
  return FALSE_OBJ;
}

void rsprof_collect_objects( obj setup, obj otbl )
{
  rsprof_start(setup);
}

#else
#include <sys/time.h>

int rsprof_active = 0;
static FILE *proff = NULL;

static char *myprof_buff, *myprof_buff_ptr, *myprof_buff_lim;
#define MYPROF_BUFF_SIZE (1024*1024)
static char path[1024];

void rsprof_cal_start( void );
void rsprof_cal_stop( void );

/* control items */

void rsprof_start( obj setup )
{
  if (STRING_P(setup))
    {
      if (proff)
	fclose(proff);
      proff = fopen( string_text(setup), "w" );
      if (!proff)
	os_error( "fopen", 1, setup );
      rsprof_active = 1;

      myprof_buff = malloc( MYPROF_BUFF_SIZE + 128 );
      myprof_buff_ptr = myprof_buff;
      myprof_buff_lim = myprof_buff + MYPROF_BUFF_SIZE;
      rsprof_cal_start();
      rsprof_cal_stop();
    }
  else
    scheme_error( "rsprof_start: argument ~s not a pathname string", 1, setup);
}

/*
 *  scans a file, looking for objects that it would be nice to have
 *  named in the output.  In particular, <<class>> and <template> objects.
 *
 *  Note that it is up to the user to guarantee that any classes and
 *  templates used during the tracing run are still around!
 */

void rsprof_collect_objects( obj setup, obj otbl )
{
  FILE *f;
  char temp[40];
  obj tmp_o;

  if (STRING_P(setup))
    {
      f = fopen( string_text(setup), "r" );
      if (!f)
	os_error( "fopen", 1, setup );
      while (1)
	{
	switch (getc(f))
	  {
	  case EOF:
	    goto done;
	  case RSPROF_MT_RETURNS:
	  case RSPROF_MT_BJUMPS:
	  case RSPROF_MT_JUMPS:
	  case RSPROF_MT_FAILS:
	  case RSPROF_MT_INTR:
	  case RSPROF_SAVES:
	  default:
	    break;
	    
	  case RSPROF_MT_CALLS:
	    fread( &tmp_o, 1, sizeof(obj), f );
	    objecttable_insert( otbl, obj_hash(tmp_o), tmp_o, tmp_o );
	    fread( &temp, 1, sizeof(UINT_32), f );
	    break;
	    
	  case RSPROF_MT_START:
	    fread( &tmp_o, 1, sizeof(obj), f );
	    fread( &temp, 1, sizeof(struct timeval), f );
	    objecttable_insert( otbl, obj_hash(tmp_o), tmp_o, tmp_o );
	    break;

	  case RSPROF_OBJ_ALLOCED:
	    fread( &tmp_o, 1, sizeof(obj), f );
	    fread( &tmp_o, 1, sizeof(obj), f );
	    objecttable_insert( otbl, obj_hash(tmp_o), tmp_o, tmp_o );
	    fread( &temp, 1, sizeof(UINT_32), f );
	    break;
	    
	  case RSPROF_MT_DONE:
	  case RSPROF_GC_WORK:
	    fread( &temp, 1, sizeof(struct timeval), f );
	    break;
	    
	  case RSPROF_RESTORED:
	  case RSPROF_CAPTURED:
	    fread( &tmp_o, 1, sizeof(obj), f );
	    break;

	  case RSPROF_OBJ_DIED:
	    fread( &tmp_o, 1, sizeof(obj), f );
	    break;
	  }
	}
    done:
      fclose(f);
    }
  else
    scheme_error( "rsprof_collect_templates: "
		  "argument ~s not a pathname string", 1, setup );
}

static void bflush( void )
{
  size_t n = myprof_buff_ptr - myprof_buff;
  if (n != fwrite( myprof_buff, 1, n, proff ))
    os_error( "fwrite", 0 );
  myprof_buff_ptr = myprof_buff;
}

#define CHECK_FOR_FLUSH(hr) do { if ((myprof_buff_ptr+(hr))>=myprof_buff_lim)\
                                  bflush(); } while (0)


obj rsprof_stop( void )
{
  if (rsprof_active)
    {
      rsprof_cal_start();
      rsprof_cal_stop();
      bflush();
      if (fclose( proff ) != 0)
	os_error( "fclose", 0 );
      proff = NULL;
      rsprof_active = 0;
      return TRUE_OBJ;
    }
  else
    return FALSE_OBJ;
}

#define PROF_CODE_HR(hr,c) CHECK_FOR_FLUSH(hr); *myprof_buff_ptr++ = c
#define PROF_CODE(c) CHECK_FOR_FLUSH(0); *myprof_buff_ptr++ = c

/* hooks to indicate how the current monotone is being exited... */

void rsprof_mt_calls( obj proc, obj tmpl )
{
  UINT_32 n = arg_count_reg;
  PROF_CODE(RSPROF_MT_CALLS);
  memcpy( myprof_buff_ptr, &tmpl, sizeof(obj) );
  myprof_buff_ptr += sizeof(obj);
  memcpy( myprof_buff_ptr, &n, sizeof(UINT_32) );
  myprof_buff_ptr += sizeof(UINT_32);
}

void rsprof_mt_returns( void )
{
  PROF_CODE(RSPROF_MT_RETURNS);
}

void rsprof_mt_bjumps( void )
{
  PROF_CODE(RSPROF_MT_BJUMPS);
}

void rsprof_mt_jumps( void )
{
  PROF_CODE(RSPROF_MT_JUMPS);
}

void rsprof_mt_fails( void )
{
  PROF_CODE(RSPROF_MT_FAILS);
}

/* hooks to keep track of the stack state */

void rsprof_saves( void )
{
  PROF_CODE(RSPROF_SAVES);
}

void rsprof_contn_captured( obj contn )
{
  PROF_CODE(RSPROF_CAPTURED);
  memcpy( myprof_buff_ptr, &contn, sizeof(obj) );
  myprof_buff_ptr += sizeof(obj);
}

void rsprof_contn_restored( obj contn )
{
  PROF_CODE(RSPROF_RESTORED);
  memcpy( myprof_buff_ptr, &contn, sizeof(obj) );
  myprof_buff_ptr += sizeof(obj);
}

void rsprof_mt_intr( void )
{
  PROF_CODE(RSPROF_MT_INTR);
}

void rsprof_mt_start( jump_addr entry_pt )
{
  struct timeval tv;
  obj t;

  gettimeofday( &tv, NULL );

  PROF_CODE_HR(120,RSPROF_MT_START);
  t = literals_reg;
  memcpy( myprof_buff_ptr, &t, sizeof(obj) );
  myprof_buff_ptr += sizeof(obj);
  memcpy( myprof_buff_ptr, &tv, sizeof(struct timeval) );
  myprof_buff_ptr += sizeof(struct timeval);
}

static void tstamped( char opcode )
{
  struct timeval tv;
  gettimeofday( &tv, NULL );
  PROF_CODE(opcode);
  memcpy( myprof_buff_ptr, &tv, sizeof(struct timeval) );
  myprof_buff_ptr += sizeof(struct timeval);
}

void rsprof_cal_start( void )
{
  tstamped( RSPROF_CAL_START );
}

void rsprof_cal_stop( void )
{
  tstamped( RSPROF_CAL_STOP );
}

void rsprof_mt_done( void )
{
  tstamped( RSPROF_MT_DONE );
}

void rsprof_gc_work( void )
{
  tstamped( RSPROF_GC_WORK );
}

void rsprof_obj_alloced( obj item, obj obj_class, UINT_32 bytes )
{
  PROF_CODE(RSPROF_OBJ_ALLOCED);
  memcpy( myprof_buff_ptr, &item, sizeof(obj) );
  memcpy( myprof_buff_ptr + sizeof(obj), &obj_class, sizeof(obj) );
  memcpy( myprof_buff_ptr + 2*sizeof(obj), &bytes, sizeof(UINT_32) );
  myprof_buff_ptr += 2*sizeof(obj) + sizeof(UINT_32);
}

void rsprof_obj_died( obj item )
{
  PROF_CODE(RSPROF_OBJ_DIED);
  memcpy( myprof_buff_ptr, &item, sizeof(obj) );
  myprof_buff_ptr += sizeof(obj);
}


#endif
